// @flow
import INTERNALS from './../../internals';
import { generatePushID, isFunction } from './../../utils';


// TODO should be PhoneAuthCredential
// TODO import 'phone' from providers
// TODO modify phone credential file to become:
/**
 export default {
          // todo maybe make this a class
          credential(verificationId, code, writable = false) {
            const out =  {
              token: verificationId,
              secret: code,
              provider: 'phone',
              providerId: 'phone',
            };

            // similar to https://firebase.google.com/docs/reference/android/com/google/firebase/auth/PhoneAuthCredential.html#getSmsCode()
            if (writable) {
              out.setSmsCode = (code) => {
                out.secret = code;
              };
            }

            // todo maybe add a getVerificationId getter fn
            out.getSmsCode = () => {
              return out.secret;
            };

            return out;
          },
        };
 */

export default class PhoneAuthListener {
  _auth: Object;
  _reject: Function | null;
  _resolve: Function | null;
  _promise: Promise | null;
  _credential: Object | null;
  _defaultTimeout: number;
  _phoneAuthRequestKey: string;
  _events: Object;

  /**
   *
   * @param auth
   * @param phoneNumber
   */
  constructor(auth: Object, phoneNumber: string) {
    this._auth = auth;
    this._reject = null;
    this._resolve = null;
    this._promise = null;
    this._credential = null;

    this._defaultTimeout = 20000; // 20 secs
    this._phoneAuthRequestKey = generatePushID();

    this._events = {
      codeSent: `phone:auth:${this._phoneAuthRequestKey}:onCodeSent`,
      verificationFailed: `phone:auth:${this._phoneAuthRequestKey}:onVerificationFailed`,
      verificationComplete: `phone:auth:${this._phoneAuthRequestKey}:onVerificationComplete`,
      codeAutoRetrievalTimeout: `phone:auth:${this._phoneAuthRequestKey}:onCodeAutoRetrievalTimeout`,
    };

    // user functions subscribed to above events
    this._subscribers = {};

    // setup event listeners
    this._subscribeToEvents();

    // start verification flow natively
    this._auth._native.verifyPhoneNumber(
      phoneNumber,
      this._phoneAuthRequestKey,
    );
  }

  /**
   * Subscribes to all EE events on this._events
   * @private
   */
  _subscribeToEvents() {
    const events = Object.keys(this._events);

    for (let i = 0, len = events.length; i < len; i++) {
      const type = events[i];
      this._auth.once(this._events[type], this[`_${type}Handler`].bind(this, type));
    }
  }

  /* ----------------------
   --- INTERNAL LISTENERS
   ------------------------ */

  _codeSentHandler(eventName, verificationId) {
    if (!this._credential) this._credential = Phone.credential(verificationId, null, true);
    if (this._subscribers[eventName]) this._subscribers[eventName](this._credential);

    // ios has no auto retrieval so instantly resolve the promise with the credential
    if (Platform.OS === 'ios') {
      if (this._resolve) this._resolve(this._credential);
    } else {
      // TODO confirm android auto retrieval timeout isn't buggy, otherwise comment this
      // TODO back in.
      // android has auto retrieval - so either wait for it to timeout via
      // a codeAutoRetrievalTimeout event, or via our internal timeout
      // this._autoRetrievalTimeout = setTimeout(
      //   this._codeAutoRetrievalTimeoutHandler.bind(this, eventName, verificationId),
      //   this._defaultTimeout,
      // );
    }
  }

  _codeAutoRetrievalTimeoutHandler(eventName, verificationId) {
    if (!this._credential) this._credential = Phone.credential(verificationId, null, true);
    if (this._subscribers[eventName]) this._subscribers[eventName](this._credential);
    if (this._resolve) this._resolve(this._credential);
    // if (this._autoRetrievalTimeout) clearTimeout(this._autoRetrievalTimeout); // TODO if android auto retrieve buggy
  }

  _verificationCompleteHandler(eventName, credential) {
    if (this._resolve) this._resolve(credential);
    if (this._subscribers[eventName]) this._subscribers[eventName](this._credential);

    // remove leftover listeners
    this._auth.removeAllListeners(this._events.codeSent);
    this._auth.removeAllListeners(this._events.verificationFailed);
    this._auth.removeAllListeners(this._events.codeAutoRetrievalTimeout);
  }

  _verificationFailedHandler(eventName, errObject) {
    // todo convert errObject to actual js error
    if (this._reject) this._reject(errObject);
    if (this._subscribers[eventName]) this._subscribers[eventName](this._credential);

    // remove leftover listeners
    this._auth.removeAllListeners(this._events.codeSent);
    this._auth.removeAllListeners(this._events.verificationComplete);
    this._auth.removeAllListeners(this._events.codeAutoRetrievalTimeout);
  }

  /**
   * Create a new internal deferred promise, if not already created
   * @private
   */
  _promiseDeferred() {
    if (!this._promise) {
      this._promise = new Promise((resolve, reject) => {
        this._resolve = (result) => {
          // TODO if android auto retrieve buggy
          // if (this._autoRetrievalTimeout) clearTimeout(this._autoRetrievalTimeout);
          this._resolve = null;
          return resolve(result);
        };

        this._reject = (possibleError) => {
          this._reject = null;
          return reject(possibleError);
        };
      });
    }
  }

  /* -------------
   -- PUBLIC API
   --------------*/

  /**
   *
   * @param cb
   * @return {*}
   */
  onCodeSent(cb: () => string): PhoneAuthListener {
    if (!isFunction(cb)) {
      throw new Error(INTERNALS.STRINGS.ERROR_MISSING_CB('onCodeSent'));
    }

    if (!this._subscribers.codeSent) this._subscribers.codeSent = cb;
    return this;
  }

  /**
   *
   * @param cb
   */
  onCodeAutoRetrievalTimeout(cb) {
    if (Platform.OS === 'ios') return this; // do nothing on ios, no platform specific code required from end user
    if (!isFunction(cb)) {
      throw new Error(INTERNALS.STRINGS.ERROR_MISSING_CB('onCodeAutoRetrievalTimeout'));
    }

    if (!this._subscribers.codeAutoRetrievalTimeout) this._subscribers.codeAutoRetrievalTimeout = cb;
    return this;
  }

  /**
   *
   * @param cb
   * @return {*}
   */
  onVerificationComplete(cb: () => Object): PhoneAuthListener {
    if (!isFunction(cb)) {
      throw new Error(INTERNALS.STRINGS.ERROR_MISSING_CB('onVerificationComplete'));
    }

    if (!this._subscribers.verificationComplete) this._subscribers.verificationComplete = cb;
    return this;
  }

  /**
   *
   * @param cb
   * @return {*}
   */
  onVerificationFailed(cb: () => Error): PhoneAuthListener {
    if (!isFunction(cb)) {
      throw new Error(INTERNALS.STRINGS.ERROR_MISSING_CB('onVerificationFailed'));
    }

    if (!this._subscribers.verificationFailed) this._subscribers.verificationFailed = cb;
    return this;
  }

  // promises are always final on any chain - so no returning 'this' required
  then(fn) {
    this._promiseDeferred();
    return this._promise.then.bind(this._promise)(fn);
  }

  catch(fn) {
    this._promiseDeferred();
    return this._promise.catch.bind(this._promise)(fn);
  }
}


// Usage example with promise chaining
// so ultimately as a user you can be lazy and just use the final promise
// to do a basic login flow or get the final result
// however for more advanced users UI's there's all the additional listeners available

firebase.auth().verifyPhoneNumber(phoneNumber)
  .onCodeSent(phoneCredential => {
    // prompt to enter auth code, save verificationId to use with entered auth code to build phone credential
  })
  .onVerificationComplete(phoneCredential => {
    // log in or link user as required
  })
  .onVerificationFailed(error => {
    // display error
  })
  .onCodeAutoRetrievalTimeout(phoneCredential => {
    // as onCodeSent
  })
  .then((phoneCredential) => { // rough example
    if (!phoneCredential.getSmsCode()) {
      phoneCredential.setSmsCode(theUsersInputHere);
    }

    return signInWithCredential(phoneCredential);
  })
  .then((currentUser) => {
    // woop signed in,
  })
  .catch((authError) => {
    // auth error comes from either PhoneAuthListener promise
    // or signInWithCredential promise as we're chained
  });
